/***************************************************************************
 *   Copypr2 (C) 2018 by hovercraft:                                     *
 *		https://github.com/hovercraft-github                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <inttypes.h>
#include "circumgyr.h"

/*
  These macro provides quick & dirty replacement for Arduino IO functions.
  Not recommended.
  begin
*/
#define LOW 0
#define HIGH 1
#define INPUT 0
#define OUTPUT 1

//#define IOPORTA 0
#define IOPORTB ('B'-'A')
#define IOPORTC ('C'-'A')
#define IOPORTD ('D'-'A')
//#define IOPORTE ('E'-'A')
//#define IOPORTF ('F'-'A')

#define DefPinNo(PortChr, PinN) ((PortChr-'A')*8 + PinN)

#define IOBitSetX(ioport, port, bit) (ioport==IOPORT ## port)?(PORT ## port|=1<<bit):
#define IOBitClearX(ioport, port, bit) (ioport==IOPORT ## port)?(PORT ## port &=~(1<<bit)):
#define IOPortReadX(ioport, port) (ioport==IOPORT ## port)?(PIN ## port):

#define IOBitSet(ioport, bit) \
  IOBitSetX(ioport, B, bit) \
  IOBitSetX(ioport, C, bit) \
  IOBitSetX(ioport, D, bit) \
  0
#define IOBitClear(ioport, bit) \
  IOBitClearX(ioport, B, bit) \
  IOBitClearX(ioport, C, bit) \
  IOBitClearX(ioport, D, bit) \
  0
#define IOPortRead(ioport) \
  IOPortReadX(ioport, B) \
  IOPortReadX(ioport, C) \
  IOPortReadX(ioport, D) \
  0

#define digitalWrite(pin,value) ((value==LOW)?IOBitClear((pin/8),(pin-((pin/8)*8))):IOBitSet((pin/8),(pin-((pin/8)*8))))
#define digitalRead(pin) ((IOPortRead(pin/8))&(1<<(pin-((pin/8)*8))))?HIGH:LOW

/*
  Above macro provides quick & dirty replacement for Arduino IO functions.
  Not recommended.
  end
*/

//segments
#define c DefPinNo('B', 1)  //9     // PB1
#define e DefPinNo('B', 2)  //10    // PB2
#define a DefPinNo('B', 3)  //11    // PB3
#define d DefPinNo('B', 4)  //12    // PB4
#define g DefPinNo('C', 3)  //A3    // PC3
#define b DefPinNo('C', 5)  //A5    // PC5
#define p DefPinNo('D', 6)  //6     // PD6
#define f DefPinNo('D', 7)  //7     // PD7

//digits
#define d3 DefPinNo('B', 0)  //8   // PB0
#define d2 DefPinNo('C', 4)  //A4  // PC4
#define d1 DefPinNo('D', 2)  //2   // PD2
#define d4 DefPinNo('D', 5)  //5   // PD5

#define KB_Read_Pin DefPinNo('C', 0)  //A0  // PC0
#define KB_Pr1Btn a   // PB3
#define KB_Pr2Btn g  // PC3
#define KB_MasterBtn c // PB1
#define KB_Menu e       // PB2
#define KB_Plus p       // PD6
#define KB_Minus f      // PD7

#define Actuator_1 DefPinNo('C', 1)
#define OUT_Fan DefPinNo('C', 2)
#define OUT_Clock DefPinNo('D', 1)  //1 // PD1

typedef enum {
  Mode_Initial = 0,
  Mode_RunPr1,
  Mode_RunBtn,
  Mode_RunPr2,
  Mode_ChooseTimerMenu,
  Mode_SetPr1Menu,
  Mode_SetPr2Menu,
  Mode_SetWatch
} Modes;

uint8_t scan_lines[8] = {a,g,c,e,p,f,b,d};
volatile uint8_t kb_state_prev = 0;
volatile uint8_t kb_state = 0;
volatile uint8_t kb_long_press = 0;
volatile int32_t kb_press_clock_n[6] = {0,0,0,0,0,0};

volatile uint8_t mode = Mode_Initial;
volatile uint8_t prev_run_mode = Mode_Initial;

typedef enum {
  KN_Pr1Btn = 0,
  KN_Pr2Btn,
  KN_MasterBtn,
  KN_Menu,
  KN_Plus,
  KN_Minus
} KeyNums;

//other
volatile int32_t prev_clock_n = -100;
volatile int32_t clock_n = 0;
volatile uint8_t blink_cntr = 0;
volatile uint8_t blink_clock = 0;
int x = 1; //100;
double del = 45.0;
volatile int16_t pr1_time = 500;
int8_t pr1_time_digits[4] = {0,5,0,0};
int8_t pr1_edit_pos = 4;
volatile int16_t pr2_time = 1000;
int8_t pr2_time_digits[4] = {1,0,0,0};
int8_t pr2_edit_pos = 4;

volatile int16_t run_act1 = 0;
volatile int16_t run_fan = 0;
volatile int16_t run_hand = 0;

volatile uint32_t watch_clock = 0;

char last_command = 0;
int8_t watch_edit_pos = 1;
int8_t watch_time_digits[2];
bool wdt_on = false;

#define KB_RELAX_TIME 3 // in 10ms units
#define KB_LONGPRESS_TIME 50

// EEPROM address table
#define EE_Init 0
#define EE_Pr1Time_LSB 1
#define EE_Pr1Time_MSB 2
#define EE_Pr2Time_LSB 3
#define EE_Pr2Time_MSB 4

void readSettings();
void storePr1Time();
void storePr2Time();

/**
Defines circular buffer and the input character queue upon it.
*/
char g_charBuff[20];
TGirString g_input(g_charBuff, sizeof(g_charBuff));

void lightsOff()
{
  // Turn off lights
  digitalWrite(d1, LOW);
  digitalWrite(d2, LOW);
  digitalWrite(d3, LOW);
  digitalWrite(d4, LOW);
}

uint8_t scanKeys()
{

  lightsOff();

  // Deactivate every scan line
  for (uint8_t i = 0; i < 8; i++) {
    digitalWrite(scan_lines[i], LOW);
  }

  kb_state_prev = kb_state;
  char chr;

  int32_t clock_n_l;
  //cli();
  clock_n_l = clock_n;
  //sei();

  // Now activate every single scan line and monitor the read pin:
  for (uint8_t i = 0; i < 8; i++) {
    _delay_us(5);
    digitalWrite(scan_lines[i], HIGH);
    _delay_us(5);
    int read = digitalRead(KB_Read_Pin);
    if (read && (i < 6)) {
        kb_state |= _BV(i);
        if (!(kb_state_prev & _BV(i))) {
            kb_press_clock_n[i] = clock_n_l;
            switch (i)
            {
            case KN_Pr1Btn:
              chr = 'L';
              break;
            case KN_MasterBtn:
              chr = 'C';
              break;
            case KN_Pr2Btn:
              chr = 'R';
              break;
            case KN_Menu:
              chr = 'M';
              break;
            case KN_Plus:
              chr = 'U';
              break;
            case KN_Minus:
              chr = 'D';
              break;
            }
            g_input.push_front(&chr);
        } else {
          if ((clock_n_l - kb_press_clock_n[i] > KB_LONGPRESS_TIME)) {
            if (i == KN_Menu) {
              kb_press_clock_n[i] = clock_n_l;
              chr = 'S';
              g_input.push_front(&chr);
            } else if (i == KN_MasterBtn) {
              kb_press_clock_n[i] = clock_n_l;
              chr = 'H';
              g_input.push_front(&chr);
            } else if (i == KN_Plus) {
              kb_press_clock_n[i] = clock_n_l;
              chr = '+';
              g_input.push_front(&chr);
            } else if (i == KN_Minus) {
              kb_press_clock_n[i] = clock_n_l;
              chr = '-';
              g_input.push_front(&chr);
            }
          }
        }
    } else {
        if (clock_n_l - kb_press_clock_n[i] > KB_RELAX_TIME) {
            kb_state &= ~_BV(i);
            if ((kb_state_prev & _BV(i)) != 0) {
              switch (i)
              {
              case KN_Pr1Btn:
                chr = 'l';
                break;
              case KN_MasterBtn:
                chr = 'c';
                break;
              case KN_Pr2Btn:
                chr = 'r';
                break;
              case KN_Menu:
                chr = 'm';
                break;
              case KN_Plus:
                chr = 'u';
                break;
              case KN_Minus:
                chr = 'd';
                break;
              }
              g_input.push_front(&chr);
            }
        }
    }
    digitalWrite(scan_lines[i], LOW);
  }
  return kb_state;
}

void clearLEDs()
{
  digitalWrite(a, HIGH);
  digitalWrite(b, HIGH);
  digitalWrite(c, HIGH);
  digitalWrite(d, HIGH);
  digitalWrite(e, HIGH);
  digitalWrite(f, HIGH);
  digitalWrite(g, HIGH);
  digitalWrite(p, HIGH);
}

void selectDigitPos(int x)
{
  lightsOff();

  switch(x)
  {
    case 1: digitalWrite(d1, HIGH); break;
    case 2: digitalWrite(d2, HIGH); break;
    case 3: digitalWrite(d3, HIGH); break;
    default: digitalWrite(d4, HIGH); break;
  }
}

void zero()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, LOW);
  digitalWrite(e, LOW);
  digitalWrite(f, LOW);
  digitalWrite(g, HIGH);
}

void one()
{
  digitalWrite(a, HIGH);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, HIGH);
  digitalWrite(e, HIGH);
  digitalWrite(f, HIGH);
  digitalWrite(g, HIGH);
}

void two()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, HIGH);
  digitalWrite(d, LOW);
  digitalWrite(e, LOW);
  digitalWrite(f, HIGH);
  digitalWrite(g, LOW);
}

void three()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, LOW);
  digitalWrite(e, HIGH);
  digitalWrite(f, HIGH);
  digitalWrite(g, LOW);
}

void four()
{
  digitalWrite(a, HIGH);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, HIGH);
  digitalWrite(e, HIGH);
  digitalWrite(f, LOW);
  digitalWrite(g, LOW);
}

void five()
{
  digitalWrite(a, LOW);
  digitalWrite(b, HIGH);
  digitalWrite(c, LOW);
  digitalWrite(d, LOW);
  digitalWrite(e, HIGH);
  digitalWrite(f, LOW);
  digitalWrite(g, LOW);
}

void six()
{
  digitalWrite(a, HIGH);
  digitalWrite(b, HIGH);
  digitalWrite(c, LOW);
  digitalWrite(d, LOW);
  digitalWrite(e, LOW);
  digitalWrite(f, LOW);
  digitalWrite(g, LOW);
}

void seven()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, HIGH);
  digitalWrite(e, HIGH);
  digitalWrite(f, HIGH);
  digitalWrite(g, HIGH);
}

void eight()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, LOW);
  digitalWrite(e, LOW);
  digitalWrite(f, LOW);
  digitalWrite(g, LOW);
}

void nine()
{
  digitalWrite(a, LOW);
  digitalWrite(b, LOW);
  digitalWrite(c, LOW);
  digitalWrite(d, HIGH);
  digitalWrite(e, HIGH);
  digitalWrite(f, LOW);
  digitalWrite(g, LOW);
}

void dispNumber(int x)
{
  switch(x)
  {
    default: zero(); break;
    case 1: one(); break;
    case 2: two(); break;
    case 3: three(); break;
    case 4: four(); break;
    case 5: five(); break;
    case 6: six(); break;
    case 7: seven(); break;
    case 8: eight(); break;
    case 9: nine(); break;
  }
}

void dispDecPoint()
{
  digitalWrite(p, LOW);
}

void timer2start()
{
  cli(); // stop interrupts, the following sequence of operations will be atomic:
  TCNT2  = 0; // initialize counter value to 0
  TIMSK2 |= (1 << TOIE2); // enable the overflow interrupt
  TIFR2 |= _BV(OCF2B) | _BV(OCF2A) | _BV(TOV2); // clear interrupt flags
  sei(); // allow interrupts
}

void timer2stop()
{
  cli(); //stop interrupts
  TIMSK2 &= ~(1 << TOIE2); // disable the overflow interrupt
  TIFR2 |= _BV(OCF2B) | _BV(OCF2A) | _BV(TOV2); // clear interrupt flags
  sei(); //allow interrupts
}

// Overflows every 0.01 second
ISR(TIMER1_COMPA_vect)
//ISR(TIMER1_OVF_vect)
{
  clock_n++;
  blink_cntr++;
  if (blink_cntr >= 10) {
    blink_cntr = 0;
    blink_clock++;
  }
  if (clock_n%2 == 0) {
    digitalWrite(OUT_Clock, LOW);
  } else {
    digitalWrite(OUT_Clock, HIGH);
  }
  if (run_fan) {
    run_fan--;
    if (!run_fan)
      digitalWrite(OUT_Fan, LOW);
  }
  if ((mode == Mode_RunPr1) || (mode == Mode_RunPr2)) {
    if (run_act1 > 0) {
      run_act1--;
      if (!run_act1) {
        mode = Mode_Initial;
        digitalWrite(Actuator_1, LOW);
      }
    } else {
      if (mode == Mode_RunPr1) {
        run_act1 = pr1_time + 1;
      } else if (mode == Mode_RunPr2) {
        run_act1 = pr2_time + 1;
      }
      run_fan = run_act1 + 3000;
      digitalWrite(Actuator_1, HIGH);
      digitalWrite(OUT_Fan, HIGH);
    }
  } else if (mode == Mode_RunBtn) {
    run_hand++;
    if (run_hand > 2000) {
        mode = Mode_Initial;
        digitalWrite(Actuator_1, LOW);
    }
  }
  TIFR1 |= _BV(TOV1); // reset overflow interrupt flag
  wdt_reset();
}

// Overflows every 1 second
ISR(TIMER2_OVF_vect)
{
  if (clock_n < 0) {
    clock_n = 0;
    prev_clock_n = -100;
  } else {
    int32_t diff = clock_n - prev_clock_n - 100;
    if (diff < 0) {
        OSCCAL += 1;
    } else if (diff > 0) {
        OSCCAL -= 1;
    }
    prev_clock_n = clock_n;
  }
  watch_clock++;
  if (watch_clock >= ((uint32_t)24L)*3600)
    watch_clock = 0;
}

void showWatch(int8_t blink_pos = 0)
{
  int hours = watch_clock/3600;
  int minutes = (watch_clock/60)%60;
  clearLEDs();
  selectDigitPos(1);
  if ((blink_clock%4 < 2) || (blink_pos != 1)
      || (kb_state & _BV(KN_Plus))
      || (kb_state & _BV(KN_Minus))
      )
    dispNumber((hours/10)%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(2);
  if (watch_clock%2 == 0)
    dispDecPoint();
  if ((blink_clock%4 < 2) || (blink_pos != 1)
      || (kb_state & _BV(KN_Plus))
      || (kb_state & _BV(KN_Minus))
      )
    dispNumber(hours%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(3);
  if ((blink_clock%4 < 2) || (blink_pos != 2)
      || (kb_state & _BV(KN_Plus))
      || (kb_state & _BV(KN_Minus))
      )
    dispNumber((minutes/10)%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(4);
  if ((blink_clock%4 < 2) || (blink_pos != 2)
      || (kb_state & _BV(KN_Plus))
      || (kb_state & _BV(KN_Minus))
      )
    dispNumber(minutes%10);
  _delay_us(del);
}

void showHundredsOfSec(int16_t ms_x_10, int8_t blink_pos)
{
  clearLEDs();
  selectDigitPos(1);
  if ((blink_clock%4 < 2) || (blink_pos != 1))
    dispNumber((ms_x_10/1000)%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(2);
  dispDecPoint();
  if ((blink_clock%4 < 2) || (blink_pos != 2))
    dispNumber((ms_x_10/100)%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(3);
  if ((blink_clock%4 < 2) || (blink_pos != 3))
    dispNumber((ms_x_10/10)%10);
  _delay_us(del);

  clearLEDs();
  selectDigitPos(4);
  if ((blink_clock%4 < 2) || (blink_pos != 4))
    dispNumber(ms_x_10%10);
  _delay_us(del);
}

void showChooseTimerMenu()
{
    clearLEDs();
    if (blink_clock%4 < 2) {
      for (int i = 1; i < 5; i++) {
        selectDigitPos(i);
        digitalWrite(g, LOW);
        _delay_us(del);
      }
    }
}

void showPr1Unit()
{
    clearLEDs();
    selectDigitPos(1);
    dispNumber(1);
    _delay_us(del);
}

void showPr2Unit()
{
    clearLEDs();
    selectDigitPos(4);
    dispNumber(2);
    _delay_us(del);
}

void showWheel()
{
    clearLEDs();
    if (blink_clock%4 == 0) {
      selectDigitPos(2);
      digitalWrite(a, LOW);
      _delay_us(del);
      clearLEDs();
      selectDigitPos(3);
      digitalWrite(a, LOW);
      _delay_us(del);
    } else if (blink_clock%4 == 1) {
      selectDigitPos(3);
      digitalWrite(b, LOW);
      digitalWrite(c, LOW);
      _delay_us(del);
    } else if (blink_clock%4 == 2) {
      selectDigitPos(2);
      digitalWrite(d, LOW);
      _delay_us(del);
      clearLEDs();
      selectDigitPos(3);
      digitalWrite(d, LOW);
      _delay_us(del);
    } else if (blink_clock%4 == 3) {
      selectDigitPos(2);
      digitalWrite(e, LOW);
      digitalWrite(f, LOW);
      _delay_us(del);
    }
}

void readSettings(void)
{
  if (eeprom_read_byte((uint8_t*)EE_Init) != 'T') {
    storePr1Time();
    storePr2Time();
    eeprom_write_byte((uint8_t*)EE_Init, 'T'); // marks once that eeprom init is done
  }
  uint8_t val = eeprom_read_byte((uint8_t*) EE_Pr1Time_LSB);
  pr1_time = 0;
  pr1_time |= val;
  val = eeprom_read_byte((uint8_t*) EE_Pr1Time_MSB);
  pr1_time |= ((int16_t)(val)<<8);

	val = eeprom_read_byte((uint8_t*) EE_Pr2Time_LSB);
  pr2_time = 0;
  pr2_time |= val;
  val = eeprom_read_byte((uint8_t*) EE_Pr2Time_MSB);
  pr2_time |= ((int16_t)(val)<<8);
}

void storePr1Time(void)
{
  // Saving last configuration
  int8_t l_pr1_time = (uint8_t)pr1_time;
  if (eeprom_read_byte((uint8_t*) EE_Pr1Time_LSB) != l_pr1_time)
      eeprom_write_byte((uint8_t*) EE_Pr1Time_LSB, l_pr1_time);
  l_pr1_time = (uint8_t) (pr1_time >> 8);
  if (eeprom_read_byte((uint8_t*) EE_Pr1Time_MSB) != l_pr1_time)
      eeprom_write_byte((uint8_t*) EE_Pr1Time_MSB, l_pr1_time);
}

void storePr2Time(void)
{
  // Saving last configuration
  int8_t l_pr2_time = (uint8_t)pr2_time;
  if (eeprom_read_byte((uint8_t*) EE_Pr2Time_LSB) != l_pr2_time)
      eeprom_write_byte((uint8_t*) EE_Pr2Time_LSB, l_pr2_time);
  l_pr2_time = (uint8_t) (pr2_time >> 8);
  if (eeprom_read_byte((uint8_t*) EE_Pr2Time_MSB) != l_pr2_time)
      eeprom_write_byte((uint8_t*) EE_Pr2Time_MSB, l_pr2_time);
}

void splitNumber2Digits(volatile const int16_t &number, int8_t (&dig_array)[4])
{
  dig_array[0] = (number/1000)%10;
  dig_array[1] = (number/100)%10;
  dig_array[2] = (number/10)%10;
  dig_array[3] = (number)%10;
}

void loop()
{

  //PRR &= ~_BV(PRTIM1);
  if (!wdt_on && clock_n > 200) {
    wdt_on = true;
    //wdt_enable(WDTO_8S);
    wdt_enable(WDTO_500MS);
  }

  if (mode == Mode_Initial) {
    //showHundredsOfSec(OSCCAL, 0);
    //showHundredsOfSec(clock_n, 0);
    //showHundredsOfSec(TCNT1, 0);
    if ((prev_run_mode == Mode_RunBtn) && run_fan)
      showHundredsOfSec(run_hand, 0);
    else
      showWatch();
  } else if (mode == Mode_SetWatch) {
    showWatch(watch_edit_pos);
  } else if (mode == Mode_RunPr1) {
    showPr1Unit();
    showWheel();
  } else if (mode == Mode_RunPr2) {
    showPr2Unit();
    showWheel();
  } else if (mode == Mode_ChooseTimerMenu) {
    showChooseTimerMenu();
  } else if (mode == Mode_SetPr1Menu) {
    showHundredsOfSec(pr1_time, pr1_edit_pos);
  } else if (mode == Mode_SetPr2Menu) {
    showHundredsOfSec(pr2_time, pr2_edit_pos);
  } else if (mode == Mode_RunBtn) {
    showHundredsOfSec(run_hand, 0);
  }

  scanKeys();


  char *p_char;
  while ((p_char = g_input.pop_back()) != 0) {
      switch (mode)
      {
      case Mode_RunPr2:
      case Mode_RunPr1:
        switch (*p_char)
        {
        case 'L':
        case 'R':
        case 'C':
          mode = Mode_Initial;
          break;
        }
        break;
      case Mode_Initial:
        switch (*p_char)
        {
        case 'L':
          mode = Mode_RunPr1;
          prev_run_mode = Mode_RunPr1;
          break;
        case 'R':
          mode = Mode_RunPr2;
          prev_run_mode = Mode_RunPr2;
          break;
        case 'C':
          run_hand = 0;
          run_fan = 3000;
          mode = Mode_RunBtn;
          prev_run_mode = Mode_RunBtn;
          digitalWrite(OUT_Fan, HIGH);
          break;
        case 'm':
        case 'S':
          if (last_command == 0)
            last_command = *p_char;
          prev_run_mode = Mode_Initial;
          mode = Mode_ChooseTimerMenu;
          break;
        case 'u':
        case 'd':
          watch_time_digits[0] = watch_clock/3600;
          watch_time_digits[1] = (watch_clock/60)%60;
          watch_edit_pos = 1;
          mode = Mode_SetWatch;
          break;
        }
        break;
      case Mode_RunBtn:
        switch (*p_char)
        {
        case 'C':
        case 'c':
          mode = Mode_Initial;
          break;
        case 'H':
          run_fan = 3000;
          break;
        }
        break;
      case Mode_ChooseTimerMenu:
        switch (*p_char)
        {
        case 'L':
          if ((last_command == 'S') && (run_hand > 0))
            pr1_time = run_hand;
          last_command = 0;
          splitNumber2Digits(pr1_time, pr1_time_digits);
          mode = Mode_SetPr1Menu;
          break;
        case 'R':
          if ((last_command == 'S') && (run_hand > 0))
            pr2_time = run_hand;
          last_command = 0;
          splitNumber2Digits(pr2_time, pr2_time_digits);
          mode = Mode_SetPr2Menu;
          break;
        case 'C':
          mode = Mode_Initial;
          break;
        }
        break;
      case Mode_SetPr1Menu:
        switch (*p_char)
        {
        case 'L':
          pr1_edit_pos--;
          if (pr1_edit_pos < 1)
            pr1_edit_pos = 4;
          break;
        case 'R':
          pr1_edit_pos++;
          if (pr1_edit_pos > 4)
            pr1_edit_pos = 1;
          break;
        case 'U':
          {
            pr1_time_digits[pr1_edit_pos-1]++;
            if (pr1_time_digits[pr1_edit_pos-1] > 9)
              pr1_time_digits[pr1_edit_pos-1] = 0;
            pr1_time = pr1_time_digits[0]*1000+pr1_time_digits[1]*100+pr1_time_digits[2]*10+pr1_time_digits[3];
          }
          break;
        case 'D':
          {
            pr1_time_digits[pr1_edit_pos-1]--;
            if (pr1_time_digits[pr1_edit_pos-1] < 0)
              pr1_time_digits[pr1_edit_pos-1] = 9;
            pr1_time = pr1_time_digits[0]*1000+pr1_time_digits[1]*100+pr1_time_digits[2]*10+pr1_time_digits[3];
          }
          break;
        case 'm':
          storePr1Time();
          mode = Mode_Initial;
          break;
        case 'c':
          readSettings(); // Cancel - restore from EEPROM
          mode = Mode_Initial;
          break;
        }
        break;
      case Mode_SetPr2Menu:
        switch (*p_char)
        {
        case 'L':
          pr2_edit_pos--;
          if (pr2_edit_pos < 1)
            pr2_edit_pos = 4;
          break;
        case 'R':
          pr2_edit_pos++;
          if (pr2_edit_pos > 4)
            pr2_edit_pos = 1;
          break;
        case 'U':
          {
            pr2_time_digits[pr2_edit_pos-1]++;
            if (pr2_time_digits[pr2_edit_pos-1] > 9)
              pr2_time_digits[pr2_edit_pos-1] = 0;
            pr2_time = pr2_time_digits[0]*1000+pr2_time_digits[1]*100+pr2_time_digits[2]*10+pr2_time_digits[3];
          }
          break;
        case 'D':
          {
            pr2_time_digits[pr2_edit_pos-1]--;
            if (pr2_time_digits[pr2_edit_pos-1] < 0)
              pr2_time_digits[pr2_edit_pos-1] = 9;
            pr2_time = pr2_time_digits[0]*1000+pr2_time_digits[1]*100+pr2_time_digits[2]*10+pr2_time_digits[3];
          }
          break;
        case 'm':
          storePr2Time();
          mode = Mode_Initial;
          break;
        case 'c':
          readSettings(); // Cancel - restore from EEPROM
          mode = Mode_Initial;
          break;
        }
        break;
        case Mode_SetWatch:
          {
            int8_t lim = 23;
            if (watch_edit_pos > 1)
              lim = 59;
            switch (*p_char)
            {
            case 'L':
              watch_edit_pos--;
              if (watch_edit_pos < 1)
                watch_edit_pos = 2;
              break;
            case 'R':
              watch_edit_pos++;
              if (watch_edit_pos > 2)
                watch_edit_pos = 1;
              break;
            case '+':
            case 'U':
              {
                watch_time_digits[watch_edit_pos-1]++;
                if (watch_time_digits[watch_edit_pos-1] > lim)
                  watch_time_digits[watch_edit_pos-1] = 0;
                uint32_t tmp = ((uint32_t)watch_time_digits[0])*3600+((uint32_t)watch_time_digits[1])*60;
                cli();
                watch_clock = tmp;
                sei();
              }
              break;
            case '-':
            case 'D':
              {
                watch_time_digits[watch_edit_pos-1]--;
                if (watch_time_digits[watch_edit_pos-1] < 0)
                  watch_time_digits[watch_edit_pos-1] = lim;
                uint32_t tmp = ((uint32_t)watch_time_digits[0])*3600+((uint32_t)watch_time_digits[1])*60;
                cli();
                watch_clock = tmp;
                sei();
              }
              break;
            case 'm':
            case 'c':
              mode = Mode_Initial;
              break;
            }
          }
          break;
      } // Switch
  } // While
  if (mode == Mode_RunBtn) {
    digitalWrite(Actuator_1, HIGH);
  } else if (mode == Mode_Initial) {
    digitalWrite(Actuator_1, LOW);
  }
  GTCCR = 0;
}

void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));
// Function Implementation
void wdt_init(void)
{
    MCUSR = 0;
    wdt_disable();

    return;
}

int main(void)
{
  wdt_disable();
  DDRB = _BV(0) | _BV(1) | _BV(2) | _BV(3) | _BV(4);
  DDRC = _BV(1) | _BV(2) | _BV(3) | _BV(4) | _BV(5);
  DDRD = _BV(2) | _BV(3) | _BV(4) | _BV(5) | _BV(6) | _BV(7);
  /*
  pinMode(d1, OUTPUT);
  pinMode(d2, OUTPUT);
  pinMode(d3, OUTPUT);
  pinMode(d4, OUTPUT);
  pinMode(a, OUTPUT);
  pinMode(b, OUTPUT);
  pinMode(c, OUTPUT);
  pinMode(d, OUTPUT);
  pinMode(e, OUTPUT);
  pinMode(f, OUTPUT);
  pinMode(g, OUTPUT);
  pinMode(p, OUTPUT);
  pinMode(KB_Read_Pin, INPUT);

  pinMode(OUT_Fan, OUTPUT);
  pinMode(Actuator_1, OUTPUT);
  pinMode(OUT_Clock, OUTPUT);
  */

  //set timer2
  cli(); //stop interrupts
  ASSR = _BV(AS2); // enable the 32768 quartz
  TCCR2A = 0; // set entire TCCR2A register to 0
  TCCR2B = _BV(CS22) | _BV(CS20); // set prescaler to clk/128, i.e. 32768/128=256, overflows 1 times per second.
  TCNT2  = 0; //initialize counter value to 0

  TCCR1A = 0;
  TCCR1B = _BV(WGM12) | _BV(CS11) | _BV(CS10); // Timer1 CTC mode, prescaler to clk/64=8000000/64=125000Hz
  //TCCR1B = _BV(WGM12) | _BV(CS10); // Timer1 CTC mode, prescaler to clk/1
  TCCR1C = 0;
  OCR1A = 1249; // i.e. overflows 100 times per second.
  //TCNT1 = 0;

  //TIMSK1 = _BV(TOIE1); // enable overflow interrupt
  TIMSK1 = _BV(OCIE1A);
  TIFR1 = _BV(TOV1) | _BV(ICF1) | _BV(OCF1A) | _BV(OCF1B); // reset interrupt flags
  TCNT1H = 0;
  TCNT1L = 0;
  PRR &= ~_BV(PRTIM1);

  sei(); //allow interrupts
  timer2start();
  GTCCR = 0;

  readSettings();

  while (true)
    loop();
}

// Refer: https://stackoverflow.com/questions/920500/what-is-the-purpose-of-cxa-pure-virtual
extern "C" void __cxa_pure_virtual() { while (1); }
